################################################################################
####################                                        ####################
####################        DXServer Class for py2dx        ####################
####################                                        ####################
################################################################################
#
# Purpose
#    This is part of py2dx, a Python interface to the OpenDX library named DXL.
#    Most of py2dx is generated automatically by SWIG and the OpenDX header file
#    (dxl.h). This file declares a Python class that maintains the state and
#    other details of OpenDX.
# Warning
#    Only one instance of this class and OpenDX can be used by py2dx.
# History
#    Original: Randall H. Hopper
#    Modifications: Kent Eschenberg eschenbe@psc.edu
#    Organization: Pittsburgh Supercomputing Center, CMU
#    $Source: /home/eschenbe/CVSvault/DX/py2dx/Install/DXServer.py,v $
#    $Author: eschenbe $
#    $Revision: 1.6 $
#    $Date: 2003/08/06 21:12:08 $
# Copyright
#    py2dx is a (c) of Pittsburgh Supercomputing Center (PSC),
#    Carnegie Mellon University, 2003. Permission to reproduce, use and prepare
#    derivative works of py2dx for internal use is granted provided the
#    copyright and "No Warranty" statements are included with all reproductions
#    and derivative works. py2dx and derivative works may also be redistributed
#    without charge provided that the copyright and "No Warranty" statements
#    are included in all redistributions.
# Warranty
#    NO WARRANTY. py2dx IS FURNISHED ON AN "AS IS" BASIS. PITTSBURGH
#    SUPERCOMPUTING CENTER  AND CARNEGIE MELLON UNIVERSITY MAKE NO WARRANTIES OF
#    ANY KIND, EITHER EXPRESSED OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
#    LIMITED TO, WARRANTY OF FITNESS FOR PURPOSE OR MERCHANABILITY, EXCLUSIVITY
#    OF RESULTS OR RESULTS OBTAINED FROM USE OF py2dx. PITTSBURGH
#    SUPERCOMPUTING CENTER AND CARNEGIE MELLON UNIVERSITY DO NOT MAKE ANY
#    WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK OR
#    COPYRIGHT INFRINGEMENT.
# Contents
#    class DXServer
#       __init__ ....................... initialize a new instance
#       SetDXCommand ................... set command to start DX
#       SetDepth ....................... set color depth of window
#       SetHWRenderMode ................ set hardware rendering mode
#       SetModeulPath .................. set path to modules
#       SetMacroPath ................... set path to macros
#       SetMDFFile ..................... set path and name of MDF file
#       Open ........................... start DX
#       Close .......................... stop DX
#       SetSynchronizationOnOff ........ turn synchronization on/off
#       SetMessageDebuggingOnOff ....... turn message debugging on/off
#       _DXBrokenConnectionHandlerCB ... callback for broken connections
#       _DXErrorHandlerCB .............. callback for errors
#       _DXMessageHandlerCB ............ general message callback
#       SetBrokenConnectionCallback .... turn broken connection on/off
#       SetErrorHandlerOnOff ........... turn error handler on/off
#       SetMessageHandlerOnOff ......... turn message handler on/off
#       _DXFileEventCB ................. callback for file events
#       _DXWorkProcCB .................. callback for server messages
#       GetFileEventCB ................. return current file callback
#       GetWorkProcCB .................. return server message callback
#       _ResetEnv ...................... remove all DX environment variables
#       _SetEnv ........................ set relevant DX environment variables
#       SetSync ........................ call DXLSync
#       WaitForIdle .................... call WaitForDXIdle
#       GetSocket ...................... call DXLGetSocket
#       Send ........................... call DXLSend
#       BeginMacroDefinition ........... call exDXLBeginmacroDefinition
#       EndMacroDefinition ............. call exDXLEndMacroDefinition
#       GetExecutionStatus ............. call DXLGetExecutionStatus
#       ExecuteOnce .................... call DXLExecuteOnce
#       ExecuteOnChange ................ call DXLExecuteOnChange
#       ExecuteOnceNamed ............... call exDXLExecuteOnceNamed
#       ExecuteOnChangeNamed ........... call exDXLExecuteOnChangeNamed
#       EndExecution ................... call DXLEndExecution
#       EndExecuteOnChange ............. call DXLEndExecuteOnChange
#       IsExecuteOnChangeActive ........ return EOC state
#       SetValue ....................... send a value
#       SetInteger ..................... send an integer
#       SetScalar ...................... send a scalar
#       SetString ...................... send a string
#       LoadVisualProgram .............. call DXLLoadVisualProgram
#       LoacScript ..................... call exDXLLoadScript
#       SequencerCtl ................... call DXLSequencerCtl
#       SetValueHandler ................ sets the value handler
#       RemoveValueHandler ............. call DXLRemoveValueHandler
# Unsupported DXL Commands
#    conn = DXLStartDX( cmd, host )
#    conn = DXLConnectToRunningServer( port, host )
#    bool = DXLIsMessagePending( conn )
#    DXLExitDX( conn )
#    DXLHandlePendingMessages( conn )
#    DXLWaitForEvent( conn )
#    DXLInitializeXMainLoop( xt_app_ctx, conn )
#    DXLUninitializeXMainLoop( conn )
# Example
#     dx = DXServer.DXServer()
#     dx.SetDepth( depth )
#     dx.SetHWRenderMode( hw_render_mode )
#     dx.Open()
#     dx.LoadVisualProgram( "mynet.net" )
#     dx.ExecuteOnChange()
#     dx.WaitForIdle()
#      ...
#     dx.WaitForIdle()
#     dx.Close()

import os, sys, types, time, DX

class DXServer:
  """Make OpenDX look like a Python object whose member functions are simplified
  versions of the low-level DXL commands. Part ot the py2dx package.
  """
  count = 0
  
  def __init__( self ):
    assert DXServer.count == 0, "Can have only 1 instance of DXServer."
    ( self.conn, self.depth, self.hw_render_mode ) = ( None, 24, "none" )
    ( self.module_path, self.macro_path, self.mdf_file ) = [None]*3
    self.exec_on_change = 0
    self.command = DX.DXEXEC
    
  def SetDXCommand( self, command ):
    """  INPUT   : command        - Command which invokes IBM Data Explorer  """

    assert self.conn == None, "Connection is not closed."
    self.command = command

  def SetDepth( self, depth ):  
    """  INPUT   : depth          - depth of windows to display in  """

    assert self.conn == None, "Connection is not closed."
    self.depth = depth

  def SetHWRenderMode( self, hw_render_mode ):
    """  INPUT   : hw_render_mode - "none", "gl", "opengl"  """
                 
    assert hw_render_mode in [ "none", "gl","opengl" ]
    assert self.conn == None, "Connection is not closed."
    self.hw_render_mode = hw_render_mode

  def SetModulePath( self, module_path ):
    assert self.conn == None, "Connection is not closed."
    self.module_path = module_path

  def SetMacroPath( self, macro_path ):
    assert self.conn == None, "Connection is not closed."
    self.macro_path = macro_path

  def SetMDFFile( self, mdf_file ):
    assert self.conn == None, "Connection is not closed."
    self.mdf_file = mdf_file

  def Open( self ):
    assert self.conn == None, "Connection is not closed."

    self._ResetEnv()                # Get rid of all DX env var settings
    self._SetEnv()                  # And set only those we want

    # Build DX invocation command
    cmd = self.command + " -execonly"
    if self.hw_render_mode != "none": 
      cmd = cmd + " -hwrender %s" % self.hw_render_mode
    if self.depth == 8 : cmd = cmd + " -8bitcmap private"
    if self.mdf_file   : cmd = cmd + (" -mdf %s" % self.mdf_file)
    if self.module_path: cmd = cmd + (" -modules %s" % self.module_path)
    if self.macro_path : cmd = cmd + (" -macros %s" % self.macro_path)

    # Fire up DX
    #   FIXME:  Allow host specification someday

    self.conn = DX.DXLStartDX( cmd, None )

    # By default, print errors and exit the program when they occur.
    #   Necessary because DX may be a zombie after an error occurs.
    #   (FIXME: One day, consider killing and restarting DX.  [...but
    #   then there's all that state to recreate.]
    self.SetErrorHandlerOnOff( 1 )

    # By default, print a message and exit if DX dies
    self.SetBrokenConnectionHandlerOnOff( 1 )

  def Close( self ):
    assert self.conn != None, "Connection is not open."
    DX.DXLExitDX( self.conn )
    self.exec_on_change = 0

  def SetSynchronizationOnOff( self, enable ):
    """  Convenience wrapper for DXLSetSynchronization.  """
    
    # This calls a our version of the routine that works, unlike IBMs.
    #   It allows us to get error callbacks right after we commit the error
    DX.DXLSetSynchronization( self.conn, enable )

  def SetMessageDebuggingOnOff( self, enable ):
    """  Convenience wrapper for DXLSetMessageDebugging.  """
    DX.DXLSetMessageDebugging( self.conn, enable )

  def _DXBrokenConnectionHandlerCB( self, conn ):
    """  When enabled, called to handle DXLSetBrokenConnectionHandler()
         callback messages.
         This version just raises an exception.
         Enable error handler invocation via SetErrorHandlerOnOff()

         Override this in superclass if you want another broken connection
         handler.
    """

    raise DX.DXError, "DX Connection Broken"

  def _DXErrorHandlerCB( self, conn, msg ):
    """  When enabled, called to handle DXLSetErrorHandler() callback messages.
         This version just prints the error to stdout and raises an exception.
         Enable error handler invocation via SetErrorHandlerOnOff()

         Error events == ERROR messages.  Both should be treated the same.
         When one is received, an exception should be thrown.  Then DX
         should be killed/restarted or just terminated because DX may be a
         zombie.  See comments in DXL_additions.i::WaitForDXIdle() for details.

         Override this in superclass if you want another error handler.
    """

    # NOTE:  Don't call any DXL routines in here.  If we're here because
    #   of a syntax error, DX is hosed and you could trigger another
    #   syntax error (even if your DXL calls are valid), which would
    #   cause this rtn to be called again.

    if self.error_handler_on:
      print "  <DXErr>       : %s" % msg

      # FIXME:
      # NOTE: Sadly, some messages sent to the DX error handler aren't
      #       crash-and-burn errors while some are.  This ugliness attempts
      #       to differentiate so we don't nuke the client needslessly.
      # EXAMPLES:
      #   "WARNING:  color map does not fully encompass data"

      if( msg[:8] != "WARNING:" ):
        raise DX.DXError, "Server Error: %s" % msg
      
  def _DXMessageHandlerCB( self, conn, msg, packet_type ):
    """  When enabled, default handler for received all DXLSetMessageHandler
         messages.
         This version just prints the messages to stdout...almost.

         If you plug in an ErrorHandler, and then you plug in an ERROR
         MessageHandler, it OVERRIDES the ErrorHandler (the ErrorHandler
         never gets called.  We fix this DXL annoyance by transferring all
         ERROR messages over to the error handler for processing.

         Error events == ERROR messages.  Both should be treated the same.
         When one is received, an exception should be thrown.  Then DX
         should be killed/restarted or just terminated because DX may be a
         zombie.  See comments in DXL_additions.i::WaitForDXIdle() for details.

         Enable message handler invocation via SetMessageHandlerOnOff()

         Override this in superclass if you want another message handler
    """

    msg_to_str_lookup = { DX.PACK_INTERRUPT  : "INTERRUPT ",
                          DX.PACK_MACRODEF   : "MACRODEF  ",
                          DX.PACK_FOREGROUND : "FOREGROUND",
                          DX.PACK_BACKGROUND : "BACKGROUND",
                          DX.PACK_ERROR      : "ERROR     ",
                          DX.PACK_MESSAGE    : "MESSAGE   ",
                          DX.PACK_INFO       : "INFO      ",
                          DX.PACK_LINQUIRY   : "LINQUIRY  ",
                          DX.PACK_LRESPONSE  : "LRESPONSE ",
                          DX.PACK_COMPLETE   : "COMPLETE  ",
                          DX.PACK_LINK       : "LINK      " }

    # Hand off all ERROR events to the error handler.
    if packet_type == DX.PACK_ERROR:
      print "Got Error via Message Handler"
      sys.stdout.flush()
      self._DXErrorHandlerCB( conn, msg )
    else:
      print "<DX Msg> %s:  %s" % ( msg_to_str_lookup[ packet_type ], msg )
      #global time1
      #try:              t = time.time() - time1
      #except NameError: (t,time1) = (0,time.time())
      #print "%8.2f: <DX Msg> %s:  %s" % \
      #       ( t, msg_to_str_lookup[ packet_type ], msg )

  def SetBrokenConnectionHandlerOnOff( self, enable ):
    if ( enable ): handler = self._DXBrokenConnectionHandlerCB
    else:          handler = None
    DX.DXLSetBrokenConnectionCallback( self.conn, handler );  

  def SetErrorHandlerOnOff( self, enable ):

    # Always plug in a callback (in DXL, there is no "unplug" API)
    DX.DXLSetErrorHandler( self.conn, self._DXErrorHandlerCB )
    self.error_handler_on = enable

  def SetMessageHandlerOnOff( self, enable, packet_types = None ):
    """  Used to enable/disable calling _DXErrorHandlerCB for error messages.

         NOTE: On new connections, a default PACK_ERROR handler is plugged in.
               But we can unplug it if we want, or plug something else in
               instead.

         ALSO NOTE:  Because of this, if you plug in an ErrorHandler, and
               then you plug in an ERROR MessageHandler, it OVERRIDES the
               ErrorHandler (the ErrorHandler never gets called.  We
               fix this DXL annoyance by transferring all ERROR
               messages over to the error handler for processing.

         INPUTS:  enable       - whether to enable or disable callbacks
                  packet_types - a list of the packet types to enable/disable
                                 printing messages for (if None, then
                                 enable/disable for all message types)
    """

    err_unplugged = 0
    
    def BuildCB( self, type ):
      def _( conn, msg, self=self, type=type ):
        self._DXMessageHandlerCB( conn, msg, type )
      return _

    if packet_types == None:
      packet_types = (
        DX.PACK_INTERRUPT,
        DX.PACK_MACRODEF,
        DX.PACK_FOREGROUND,
        DX.PACK_BACKGROUND,
        DX.PACK_ERROR,
        DX.PACK_MESSAGE,
        DX.PACK_INFO,
        DX.PACK_LINQUIRY,
        DX.PACK_LRESPONSE,
        DX.PACK_COMPLETE,
        DX.PACK_LINK
     )

    for i in packet_types:
      if enable:
        DX.DXLSetMessageHandler( self.conn, i, None, BuildCB( self, i ) )
      else:
        DX.DXLRemoveMessageHandler( self.conn, i, None )
        if i == DX.PACK_ERROR:
          err_unplugged = 1

    # Make sure after it's all said and done that "some" handler is still
    #   plugged in to handle errors.  This is important!
    if err_unplugged:
      self.SetErrorHandlerOnOff(1)

  def _DXFileEventCB( self ):
    """ This callback should be invoked when activity is detected on the file
        descriptor associated with our DX connection.  The client
        determines when it is called.

        If you're familiar with XtAddInput() callbacks in Xt, this routine
        serves the same function.  Essentially when the inner select() of
        the application's main event loop says there is data available on
        the DX socket, call this routine to process it.

        E.g. For Tkinter, this routine can be hooked into the GUI's
        select() in its inner mainloop via createfilehandler().

        DX descriptor data is available when we're called, so process all
        pending DX messages and return.

        FIXME: Look at performance of this routine vs. frequency with which
               it is called and, if needed, consider implementing this in
               C like WaitForDXIdle

        INPUT  : file - a Python file for the DX connection
                 mask - the mask used when registering the file
                        (e.g. Tkinter.READABLE)

    """
    DX.DXLHandlePendingMessages( self.conn );

  def _DXWorkProcCB( self ):
    """ This callback should be invoked periodically to process DX server
        messages queued but not yet delivered.

        Basically when DX receives a message from the server, it parses it,
        stuffs it into a structure, and puts it on an event queue
        (see _dxl_ReadFromSocket).

        In some circumstances, the events are all immediately processed
        (e.g. DXLHandlePendingMessages), invoking user callbacks as needed.
        However, in other scenarios, the events are queued but NOT
        processed (DXLSendPacket, DXLNextPacket, DXLGetPacket,
        DXLPeekPacket).

        Therefore we cannot simply use file descriptor activity on the DX
        socket descriptor as our queue to wake up and
        HandlePendingMessages.  This routine should be called periodically
        to process any 'queued but not processed' messages from the DX
        executive. 
       
        For more details, see the DXL implementation of InitializeXMainLoop
        and HandlePendingMessages.
    """
    DX.DXLHandlePendingMessages( self.conn );

  def GetFileEventCB( self ):
    """ Returns a callable object taking no arguments to be invoked
        when descriptor activity is detected on the DX server socket.

        It cleans out the socket input buffer, parses and dispatches all
        available events, and then returns.

        RETURNS: callable object
    """

    return self._DXFileEventCB

  def GetWorkProcCB( self ):
    """ Returns a callable object taking no arguments to be invoked
        when periodically to dispatch 'queued but not processed' events in
        the DXL client layer.

        It cleans out the socket input buffer, parses and dispatches all
        available events, and then returns.

        RETURNS: callable object
    """

    return self._DXWorkProcCB

  def _ResetEnv( self ):
    """  Remove all DX variables from the environment.  DX behavior should
         be fully specified here and in the DX invocation script
         DXLStartDX invokes.
    """
    for var in os.environ.keys():
      if var[:2] == 'DX':
        del os.environ[ var ]

  def _SetEnv( self ):
    """  Set the DX env vars we want to be in effect for our DX executive.  """

    os.environ[ 'DXNO_BACKING_STORE' ] = "1"  # Make any difference??
    os.environ[ 'DXFLING'            ] = "1"  # Keep rotating after release

  #
  # Pass-through methods
  #
  def Sync( self ): DX.DXLSync( self.conn )
  def WaitForIdle( self ): DX.WaitForDXIdle( self.conn )
  def GetSocket  ( self ): return DX.DXLGetSocket( self.conn )
  def Send  ( self, msg ): DX.DXLSend( self.conn, msg )
  def BeginMacroDefinition( self, mhdr ):
    DX.exDXLBeginMacroDefinition( self.conn, mhdr )
  def EndMacroDefinition( self ):
    DX.exDXLEndMacroDefinition( self.conn )

  def GetExecutionStatus( self ): return DX.DXLGetExecutionStatus( self.conn )
  def ExecuteOnce       ( self ):
    self.exec_on_change = 0
    DX.DXLExecuteOnce    ( self.conn )
  def ExecuteOnChange   ( self ):
    self.exec_on_change = 1
    DX.DXLExecuteOnChange( self.conn )
  def ExecuteOnceNamed    ( self, name ):
    self.exec_on_change = 0
    DX.exDXLExecuteOnceNamed( self.conn, name )
  def ExecuteOnChangeNamed( self, name ):
    self.exec_on_change = 1
    DX.exDXLExecuteOnChangeNamed( self.conn, name )
  def EndExecution      ( self ):
    self.exec_on_change = 0
    DX.DXLEndExecution( self.conn )
  def EndExecuteOnChange( self ):
    self.exec_on_change = 0
    DX.DXLEndExecuteOnChange( self.conn )

  # And a new state query
  def IsExecuteOnChangeActive( self ):
    return self.exec_on_change

  # APIs to set DXLInput and DXLInputNamed values
  #   (see the DXLInput note at bottom of file for why DXLSend is here too)
  def SetValue  ( self, var, value ):
    assert type(value) == types.StringType and value
    DX.DXLSetValue  ( self.conn, var, value )
    DX.DXLSend      ( self.conn, '%s = %s;' % (var,value) )
  def SetInteger( self, var, value ):
    DX.DXLSetInteger( self.conn, var, value )
    DX.DXLSend      ( self.conn, '%s = %s;' % (var,value) )
  def SetScalar ( self, var, value ):
    DX.DXLSetScalar ( self.conn, var, value )
    DX.DXLSend      ( self.conn, '%s = %s;' % (var,value) )
  def SetString ( self, var, value ):
    assert type(value) == types.StringType and value
    DX.DXLSetString ( self.conn, var, value )
    DX.DXLSend      ( self.conn, '%s = "%s";' % (var,value) )

  def LoadVisualProgram ( self, file ): DX.DXLLoadVisualProgram(self.conn,file)
  def LoadScript        ( self, file ): DX.exDXLLoadScript( self.conn, file )

  def SequencerCtl( self, action ):
    DX.DXLSequencerCtl( self.conn, action )

  #
  # Callback Registration Pass-through methods - 
  #   makes the first argument to the callback this DXServer instance rather
  #   than a DXLConnection.
  #
  def SetValueHandler( self, name, callback ):
    def BuildCB( self, callback ):
      def _( conn, name, value, callback=callback, self=self ):
        callback( self, name, value )
      return _

    DX.DXLSetValueHandler( self.conn, name, BuildCB( self, callback ) )
    
  def RemoveValueHandler( self, name ):
    DX.DXLRemoveValueHandler( self.conn, name )

  # NOT IMPLEMENTED FOR NOW (no need)
  def ExecuteOnceNamedWithArgs     ( self, name, *args ):
    raise NotImplementedError
  def ExecuteOnceNamedWithArgsV    ( self, name, args ):
    raise NotImplementedError
  def ExecuteOnChangeNamedWithArgs ( self, name, *args ):
    raise NotImplementedError
  def ExecuteOnChangeNamedWithArgsV( self, name, args ):
    raise NotImplementedError

  def Query( self, msg ):
    """ What would the client do with this?  DXL uses DXLQuery internally
        for DXLSync, DXLGetValue, and DXLGetExecutionStatus.
    """
    raise NotImplementedError

#############################  End of DXServer.py  #############################
