#!/usr/bin/env python
#
#  7-hwrender.py  -  Test of DX rendering in a Tk window using GL, OpenGL,
#                    and no hardware acceleration at all.

import sys, os, time, math, socket, string, DX
from Tkinter import *

IMAGE_WIDTH  = 800
IMAGE_HEIGHT = 600
NUM_STEPS    = 100

#-----------------------------------------------------------------------------

def BuildColoredBox( conn ):
  """  Generate a colored box to use for display in DX.

       INPUT  :  conn     - open DX connection
       RETURNS:  <string> - name of DX script language variable holding object
  """
  
  DX.DXLSend( conn, 'grid = Grid( point=[0,0,0], structure="brick",' +
                         'shape={[1,0,0],[0,1,0],[0,0,1]}, density={2,7,2} );' )
  DX.DXLSend( conn, 'grid2 = Mark( grid, name="positions" );' )
  DX.DXLSend( conn, 'grid3 = Compute( "$0.y", grid2 );' )
  DX.DXLSend( conn, 'grid4 = AutoColor( grid3 );' )
  DX.DXLSend( conn, 'box   = ShowBoundary( grid4 );' )
  return 'box'

#-----------------------------------------------------------------------------

def FrameBlast( picture, hw_render_mode, rendering_approx ):
  """  Display images in the window owned by a Tk widget (picture).

       INPUTS:  hw_render_mode = { "none", "gl", "opengl" }
  """

  assert hw_render_mode in [ "none", "gl","opengl" ]
  if hw_render_mode != "none":
    assert rendering_approx in [ "none", "box", "dots", "wireframe" ]
  else:
    rendering_approx = ""

  depth = picture.winfo_depth()

  print "\nPERFORMING DISPLAY TEST:"
  print "  Hardware Acceleration  : %s" % hw_render_mode
  print "  Rendering Approximation: %s" % rendering_approx
  if hw_render_mode == "none":
    print "  Depth                  : %d" % depth

  os.putenv( "DXNO_BACKING_STORE", "1" );     # Make any difference??
  
  cmd = DX.DXEXEC + " -execonly"
  if hw_render_mode != "none": cmd = cmd + " -hwrender %s" % hw_render_mode
  if depth          == 8     : cmd = cmd + " -8bitcmap private"
  conn = DX.DXLStartDX( cmd, None )

  # Create a colored box
  obj_tag = BuildColoredBox( conn )

  if hw_render_mode == "none":
    DX.DXLSend( conn, 'box2  = %s;' % obj_tag )
  else:
    # This little gem is needed to disable HW-render-associated mouse bindings
    DX.DXLSend( conn, 'iact_mode = CollectNamed( 1, "mode" );' )   # 1 = NONE
      
    DX.DXLSend( conn, ('box2  = Options( %s, "rendering mode", "hardware", ' +
                    '"rendering approximation", "%s", ' +
                    '"interaction mode", iact_mode );') %
                   ( obj_tag, rendering_approx ) )

  where_str = 'X%d,,#X%d' % ( picture.winfo_depth(), picture.winfo_id() )

  # Create a macro to display rotated views of the object
  DX.exDXLBeginMacroDefinition( conn, 'macro SHOWBOX( look_to_v )' )
  DX.DXLSend( conn, ('camera = AutoCamera( box2, resolution=%d, ' +
                  'aspect = %f, direction = look_to_v );') %
                 ( IMAGE_WIDTH, float(IMAGE_HEIGHT)/IMAGE_WIDTH ) )
  DX.DXLSend( conn, 'Display( box2, camera, where="%s" );' % where_str )
  DX.exDXLEndMacroDefinition( conn )

  #  Now benchmark frame blasting for 10 seconds  */
  time1 = time2 = time.time()
  num_frames = 0
  step       = 0

  while time2 - time1 < 10.0:

    angle = (float(step) / NUM_STEPS) * 2 * math.pi
    (x,y,z) = ( math.cos(angle), 0.5, math.sin(angle) )

    # Create a "main" macro.  We have a choice: call the macro something
    # else and use exDXLExecuteOnceNamed, or call it main() and use
    # DXLExecuteOnce.   You don't want to call it something else and use
    # DXLExecuteOnce because then DX calls an error handler twice and eats
    # socket and CPU bandwidth in two processes (not to mention making the
    # error handler useless).
    DX.exDXLBeginMacroDefinition( conn, 'macro main()' )
    DX.DXLSend( conn, 'SHOWBOX( [%f,%f,%f] );' % (x,y,z) )
    DX.exDXLEndMacroDefinition( conn )

    DX.DXLExecuteOnce( conn )
    DX.WaitForDXIdle( conn )
    # DXLSync( conn )                    # Force DX to XSync()

    num_frames = num_frames + 1
    step = ( step + 1 ) % NUM_STEPS
    time2 = time.time()


  DX.WaitForDXIdle( conn )
  DX.DXLSync( conn )                    # Force DX to XSync()

  elapsed_sec = ( time2 - time1 )

  print "  %d frames displayed in %.2f seconds ==> %.2f frames per second" \
        % ( num_frames, elapsed_sec, num_frames / elapsed_sec )

  # FIXME:  Strange.  The above loop requires some DX interaction
  #                   via WaitForDXIdle to work.  But the below loop doesn't.
  #                    WHY!?
  #
  ## # Display rotated views of the object for 10 seconds
  ## for step in range( 100 ):
  ##   angle = (step / 100.0) * 2 * math.pi
  ##   (x,y,z) = ( math.cos(angle), 0.5, math.sin(angle) )
  ##   DXLSend( conn, "SHOWBOX( [%f,%f,%f] );" % (x,y,z) )
  ##   DXLExecuteOnce( conn )
  ##   # WaitForDXIdle( conn )
  ##   # DXLSync( conn )                    # Force DX to XSync()

  DX.DXLExitDX( conn )

#-----------------------------------------------------------------------------

if __name__ == "__main__":

  depth = 0

  # Parse arguments
  if len( sys.argv ) > 1:
    try:    depth = int( sys.argv[2] )
    except: depth = 0
    if depth not in [ 8, 24 ] or sys.argv[1] != "-depth":
      print "FORMAT: %s [-depth 8|24]\n" % sys.argv[0]
      sys.exit(0)
     
  # Determine the depth, if non specified
  root     = Tk()
  root.wm_withdraw()

  if not depth: depth = root.winfo_depth()

  # Just force 24-bit
  depth = 24

  if depth == 8: visual = "pseudocolor 8"
  else:          visual = "truecolor  24"

  # Create the GUI
  main     = Toplevel( root, visual=visual )
  frame    = Frame( main, visual=visual, 
                    width = IMAGE_WIDTH, height = IMAGE_HEIGHT )
  frame.pack_propagate( 0 )

  #  NOTE:  When DX quits and we're doing hardware rendering, the window flips
  #    out of double/depth buffer and we see aftereffects from the hidden
  #    buffer if we don't clear the window
  picture  = Frame( frame, background="", colormap="new" )
  button   = Button( main, text="Quit", command = main.destroy )

  # main.wm_colormapwindows( picture, frame )

  picture.pack( expand=Y, fill=BOTH )
  frame.pack( expand=N )
  button.pack()

  main.update_idletasks()

  # Now run display benchmarks in a Tk X Window
  localhost   = socket.gethostname()
  displayhost = string.split( root.winfo_screen(), ':' )[0]
  if not displayhost or displayhost == "unix":
    displayhost = localhost
  
  print "\nPYTHON-DX COLOR CUBE BENCHMARK"
  print   "______________________________\n"
  print "Local Machine: %s" % localhost
  print "Display      : %s" % displayhost

  FrameBlast( picture, "opengl", "none"      )
  FrameBlast( picture, "gl"    , "none"      )
  FrameBlast( picture, "none"  , "none"      )
  FrameBlast( picture, "opengl", "box"       )
  FrameBlast( picture, "opengl", "wireframe" )
  FrameBlast( picture, "opengl", "dots"      )
