#!/usr/bin/env python
#
#  8-userctl.py  -  Now give the user control of DX rendering in a Tk window
#                   via mouse translations.
#
#                   In this case, the mouse events are grabbed by the Python
#                   app and used to change the camera via DXLSend.
#
#  NOTES : 1) You'll need to run DX with the attached patch for this to work
#             with hardware-rendering (DX otherwise steals Button events).
#          2) With DX as-is, this approach supports software-rendered windows
#             but not hardware (without the DX patch)
#
#  FIXME:  Every once in a while this thing will just freeze for no
#          apparent reason.  dxexec's CPU usage goes through the roof and
#          stays there for a while.
  

import sys, os, math, socket, string, DX
from Tkinter import *

IMAGE_WIDTH      = 640
IMAGE_HEIGHT     = 480
HW_RENDER_MODE   = "opengl"                # "none", "gl", "opengl" 
#HW_RENDER_MODE   = "none"                # "none", "gl", "opengl" 
HW_RENDER_APPROX = "none"                  # "none", "box", "dots", "wireframe"
FOV_ANGLE        = 30

In_drag = 0

#-----------------------------------------------------------------------------

def ROTATE( v, angle_deg, axis ):
  assert axis in [0,1,2]

  def DEG_TO_RAD(x): return x * math.pi / 180.0

  angle = DEG_TO_RAD( angle_deg )
  ( cos_angle, sin_angle ) = ( math.cos( angle ), math.sin( angle ) )

  if axis == 0:   v = (  v[0],
                         v[1] * cos_angle - v[2] * sin_angle,
                         v[1] * sin_angle + v[2] * cos_angle )

  elif axis == 1: v = (  v[0] * cos_angle + v[2] * sin_angle,
                         v[1],
                        -v[0] * sin_angle + v[2] * cos_angle )

  else:           v = (  v[0] * cos_angle - v[1] * sin_angle,
                         v[0] * sin_angle + v[1] * cos_angle,
                         v[2] )

  return v
          
#-----------------------------------------------------------------------------

class Box:
  """  A colored box object displayed in DX.  """
  
  def __init__( self, conn, wgt, hw_render_mode, hw_render_approx ):
    """  Generate a colored box to use for display in DX.

         INPUT  :  conn             - open DX connection
                   wgt              - widget to tell DX to blast in
                   hw_render_mode   - "none", "gl", "opengl"
                   hw_render_approx - "none", "box", "dots", "wireframe"
                   
         RETURNS:  <string> - name of DX script language variable holding obj
    """

    assert hw_render_mode   in [ "none", "gl","opengl" ]
    assert hw_render_approx in [ "none", "box", "dots", "wireframe" ]

    self.conn             = conn
    self.wgt              = wgt
    self.hw_render_mode   = hw_render_mode
    self.hw_render_approx = hw_render_approx

    DX.DXLSend( conn, 'grid = Grid( point=[0,0,0], structure="brick",' +
                           'shape={[1,0,0],[0,1,0],[0,0,1]}, density={2,7,2} );' )
    DX.DXLSend( conn, 'grid2 = Mark( grid, name="positions" );' )
    DX.DXLSend( conn, 'grid3 = Compute( "$0.y", grid2 );' )
    DX.DXLSend( conn, 'grid4 = AutoColor( grid3 );' )
    DX.DXLSend( conn, 'box   = ShowBoundary( grid4 );' )

    if hw_render_mode == "none":
      DX.DXLSend( conn, 'box2  = box;' )
    else:
      # This little gem is needed to disable HW-render-associated mouse bindings
      #
      # FIXME:  Even though it disables the HW rotate interactor, DX still
      #   steals all our ButtonPress/ButtonRelease events (OpenDX 4.0.9 and
      #   commercial DX 3.1.4B).  You'll need the attached OpenDX patch
      #   to fix this.
      #
      DX.DXLSend( conn, 'iact_mode = CollectNamed( 1, "mode" );' )    # 1 = NONE
      
      DX.DXLSend( conn, ('box2  = Options( box, "rendering mode", "hardware", ' +
                      '"rendering approximation", "%s", ' +
                      '"interaction mode", iact_mode );') % 
                     hw_render_approx )

    # Create a macro to display rotated views of the object
    ( width, height ) = ( wgt.winfo_width(), wgt.winfo_height() )
    where_str         = 'X%d,,#X%d' % \
                        ( wgt.winfo_depth(), wgt.winfo_id() )
    
    DX.exDXLBeginMacroDefinition( conn, 'macro SHOWBOX( look_to_v, up_v )' )
    DX.DXLSend( conn, ('camera = AutoCamera( box2, resolution=%d, ' +
                    'aspect = %f, angle = %f, direction = look_to_v, ' +
                    'up = up_v );') %
                   ( width, float( height ) / width, FOV_ANGLE ) )
    DX.DXLSend( conn, 'Display( box2, camera, where="%s" );' % where_str )
    DX.exDXLEndMacroDefinition( conn )

    self.display_macro = 'SHOWBOX'
    self.orient_v      = ROTATE( ROTATE( [ 1.0, 0.0, 0.0 ], 30, 2 ),
                                 30, 1 )
    self.up_v          = ROTATE( ROTATE( [ 0.0, 1.0, 0.0 ], 30, 2 ),
                                 30, 1 )

  def SetOrientation( self, orient_v, up_v ):

    self.orient_v = orient_v[:]
    self.up_v     = up_v    [:]

  def Display( self ):

    # Create a "main" macro.  Otherwise DX works, but calls an error
    #   handler twice and eating socket and CPU bandwidth in two processes,
    #   not to mention making the error handler useless.
    DX.exDXLBeginMacroDefinition( self.conn, 'macro main()' )
    DX.DXLSend( self.conn, '%s( [%f,%f,%f], [%f,%f,%f] );' %
                   ( (self.display_macro,) + tuple(self.orient_v)
                                           + tuple(self.up_v) ))
    DX.exDXLEndMacroDefinition( self.conn )
    DX.DXLExecuteOnce( self.conn )
    DX.WaitForDXIdle( self.conn )
    # DXLSync( self.conn )                    # Force DX to XSync()

#-----------------------------------------------------------------------------

def InitDX( depth, hw_render_mode ):
  """  Initializes and returns a connection to the DX executive.

       INPUT   : hw_render_mode - "none", "gl", "opengl"
                 depth          - 8 or 24; depth of windows to display in
       RETURNS : conn           - an open DX connection

  """

  assert hw_render_mode in [ "none", "gl","opengl" ]

  os.putenv( "DXNO_BACKING_STORE", "1" );     # Make any difference??
  os.putenv( "DXFLING"           , "1" );     # Rotate after release
  
  cmd = DX.DXEXEC + " -execonly"
  if hw_render_mode != "none": cmd = cmd + " -hwrender %s" % hw_render_mode
  if depth          == 8     : cmd = cmd + " -8bitcmap private"
  conn = DX.DXLStartDX( cmd, None )

  return conn

#-----------------------------------------------------------------------------

def test():

  if HW_RENDER_MODE != "none":
    print "\nNOTE: DX/OpenDX require a patch for this to work with " \
          "hardware rendering."
  depth = 0

  # Parse arguments

  if len( sys.argv ) > 1:
    try:    depth = int( sys.argv[2] )
    except: depth = 0
    if depth not in [ 8, 24 ] or sys.argv[1] != "-depth":
      print "FORMAT: %s [-depth 8|24]\n" % sys.argv[0]
      sys.exit(0)
     
  # Determine the depth, if non specified
  root     = Tk()
  root.wm_withdraw()

  if not depth: depth = root.winfo_depth()

  if depth == 8: visual = "pseudocolor 8"
  else:          visual = "truecolor  24"

  # Create the GUI
  main     = Toplevel( root, visual=visual )
  frame    = Frame( main, visual=visual, 
                    width = IMAGE_WIDTH, height = IMAGE_HEIGHT )
  frame.pack_propagate( 0 )

  picture  = Frame( frame, background="", colormap="new" )
  button   = Button( main, text="Quit", command = root.quit )

  # main.wm_colormapwindows( picture, frame )

  picture.pack( expand=Y, fill=BOTH )
  frame.pack( expand=N )
  button.pack()

  main.update_idletasks()

  # Display config
  localhost   = socket.gethostname()
  displayhost = string.split( root.winfo_screen(), ':' )[0]
  if not displayhost or displayhost == "unix":
    displayhost = localhost

  print "\nPYTHON-DX COLOR CUBE INTERACTION"
  print   "________________________________\n"
  print "CONFIGURATION:"
  print "  Local Machine          : %s" % localhost
  print "  Display                : %s" % displayhost
  print "  Hardware Acceleration  : %s" % HW_RENDER_MODE
  print "  Rendering Approximation: %s" % HW_RENDER_APPROX

  # Create a DX connection displaying a color cube in our Tk widget's X window
  conn  = InitDX( depth, HW_RENDER_MODE )
  box   = Box( conn, picture, HW_RENDER_MODE, HW_RENDER_APPROX )

  # Now let the user twiddle with it
  picture.bind( "<Expose>", lambda event, box=box: box.Display() )

  # FIXME:  DX is stealing all ButtonPress and Button Release events.
  #  This happens because DX (in hardware rendering mode ONLY) is binding
  #  in its own event handler with ButtonPressMask and ButtonReleaseMask set,
  #  and since only one process can get these events per window, and since X
  #  gives them to the last process registered (DX), DX steals all these events.
  #
  #  We can get all the Motion and Expose events we want since DX doesn't
  #  bind for these, but without Button events, that's not too useful.
  #
  #  NOTE:  Setting "interaction mode" to 1 (NONE) does not cause DX to
  #         change its event mask and stop eating these events as it should.
  #         This seems to be a bug.
  #
  # See below for a DX patch to fix this.
  #
  
  def start_cb( event, box=box ):
    global In_drag, Last_x, Last_y

    ( Last_x, Last_y ) = ( event.x, event.y )
    In_drag = 1
    
  def end_cb( event, box=box ):
    global In_drag, Last_x, Last_y
    
    In_drag = 0

  def drag_cb( event, box=box ):
    global In_drag, Last_x, Last_y

    assert( In_drag )

    ( dx, dy ) = ( -(event.x - Last_x), event.y - Last_y )
    ( Last_x, Last_y ) = ( event.x, event.y )

    # Heuristic
    w = event.widget.winfo_width()
    orient_v = box.orient_v[:]
    up_v     = box.up_v[:]

    angle_x = 4 * float(dx) / w * FOV_ANGLE
    #angle_y = 4 * float(dy) / w * FOV_ANGLE
    angle_y = 0.0

    orient_v = ROTATE( orient_v, angle_x, 1 )
    orient_v = ROTATE( orient_v, angle_y, 2 )
    up_v     = ROTATE( up_v    , angle_x, 1 )
    up_v     = ROTATE( up_v    , angle_y, 2 )

    box.SetOrientation( orient_v, up_v )
    box.Display()

  picture.bind( "<ButtonPress>"  , start_cb )
  picture.bind( "<ButtonRelease>", end_cb   )
  picture.bind( "<B1-Motion>"    , drag_cb  )

  root.mainloop()

  DX.DXLExitDX( conn )

if __name__ == "__main__":
  test()


"""

A PATCH FOR OPENDX 4.0.9 TO STOP THE HARDWARE RENDERER FROM STEALING
BUTTON EVENTS WHEN YOU'VE TOLD IT "NO INTERACTOR".

--- src/exec/hwrender/ORIG/hwClientMessage.c	Fri Jun 25 16:49:40 1999
+++ src/exec/hwrender/hwClientMessage.c	Tue Oct 26 14:31:28 1999
@@ -635,6 +635,25 @@
 	default:
 	    break;
     }
+
+#ifndef OLDSTUFF
+    /*  Update the window event mask  */
+    {
+        XWindowAttributes win_attr;
+        long              new_mask;
+        DEFGLOBALDATA(globals);
+
+        XGetWindowAttributes( DPY, XWINID, &win_attr );
+        new_mask = win_attr.your_event_mask;
+
+        if ( mode == INTERACTION_NONE )
+          new_mask &= ~(ButtonPressMask | ButtonReleaseMask);
+        else
+          new_mask |=  (ButtonPressMask | ButtonReleaseMask);
+
+        XSelectInput( DPY, XWINID, new_mask );
+    }
+#endif
 }
 
 Error

"""
