#!/usr/bin/env python
#
#  9-userctl2.py  -  Give the user control of DX rendering in a DX window
#                    via mouse translations.
#
#                    This time, we use SuperviseWindow and SuperviseState to
#                    let the user manipulate the cube.
#
#  NOTE:  For reasons we don't yet understand, you must encase the
#         the DXLSend-created visual program inside a macro in order for the
#         software-rendering interactors to work.  The hardware interactors
#         work regardless -- go figure!
#
  
import sys, os, socket, string, DX
from Tkinter import *

IMAGE_WIDTH      = 640
IMAGE_HEIGHT     = 480
HW_RENDER_MODE   = "opengl"                # "none", "gl", "opengl" 
#HW_RENDER_MODE   = "none"                # "none", "gl", "opengl" 
HW_RENDER_APPROX = "none"                 # "none", "box", "dots", "wireframe"

#-----------------------------------------------------------------------------

class Box:
  """  A colored box object displayed in DX.  """
  
  def __init__( self, conn, wgt, hw_render_mode, hw_render_approx ):
    """  Generate a colored box to use for display in DX.

         INPUT  :  conn             - open DX connection
                   wgt              - widget to tell DX to blast in
                   hw_render_mode   - "none", "gl", "opengl"
                   hw_render_approx - "none", "box", "dots", "wireframe"
                   
         RETURNS:  <string> - name of DX script language variable holding obj
    """

    assert hw_render_mode   in [ "none", "gl","opengl" ]
    assert hw_render_approx in [ "none", "box", "dots", "wireframe" ]

    self.conn             = conn
    self.wgt              = wgt
    self.hw_render_mode   = hw_render_mode
    self.hw_render_approx = hw_render_approx

    # BUILD: box = Grid->Mark->Compute->AutoColor->ShowBoundary

    #
    # NOTE:  For reasons we don't yet understand, you must encase the
    #  the DXLSend-created visual program inside a macro in order for the
    #  software-rendering interactors to work.  The hardware interactors
    #  work regardless -- go figure!
    #
    DX.exDXLBeginMacroDefinition( conn, 'macro main()' )

    DX.DXLSend( conn, 'grid = Grid( point=[0,0,0], structure="brick",' +
                           'shape={[1,0,0],[0,1,0],[0,0,1]}, density={2,7,2} );' )
    DX.DXLSend( conn, 'grid2 = Mark( grid, name="positions" );' )
    DX.DXLSend( conn, 'grid3 = Compute( "$0.y", grid2 );' )
    DX.DXLSend( conn, 'grid4 = AutoColor( grid3 );' )
    DX.DXLSend( conn, 'box   = ShowBoundary( grid4 );' )

    # BUILD: box2 = box->Options

    DX.DXLSend( conn, 'iact_mode = CollectNamed( 2, "mode" );' )    # 2 = ROTATE

    if hw_render_mode == "none":
      options = '"rendering mode", "software"'
    else:
      options = '"rendering mode", "hardware", ' + \
                ( '"rendering approximation", "%s"' % hw_render_approx )
    options = options + ', "interaction mode", iact_mode'

    DX.DXLSend( conn, 'box2  = Options( box, %s );' % options )

    # Create Supervise objects
    #
    # NOTE:  For HW accel mode, SuperviseState.mode seems to override
    #        Options("interaction mode"), if specified

    # BUILD: Display( box2, camera ) via SuperviseState/SuperviseWindow
    
    ( width, height, depth ) = ( wgt.winfo_width(), wgt.winfo_height(),
                                 wgt.winfo_depth() )

    DX.DXLSend( conn, ('camera = AutoCamera( box2, resolution=%d, ' +
                    'aspect = %f, direction = [1,0,0], ' +
                    'up = [0,1,0] );') %
                   ( width, float( height ) / width ) )

    DX.DXLSend( conn, ('where, size, events = SuperviseWindow( "Test", ' +
             'display = NULL, size = [ %d, %d ], offset = NULL, ' +
             'parent = %s, depth = %d );') %
             ( width, height, wgt.winfo_id(), depth ) )

    DX.DXLSend( conn, 'object, camera, where2, events = SuperviseState( where, ' +
             'defaultCamera = camera, object = box2, size = size, ' +
             'events = events, mode = -1 );' )
                 # MODES = { -1:None, 0:Rotate, 1:Pan, 2:Zoom }

    DX.DXLSend( conn, 'Display( object, camera, where=where2 );' )
    DX.exDXLEndMacroDefinition( conn )

    DX.DXLExecuteOnChange( conn )
      
#-----------------------------------------------------------------------------

def InitDX( depth, hw_render_mode ):
  """  Initializes and returns a connection to the DX executive.

       INPUT   : hw_render_mode - "none", "gl", "opengl"
                 depth          - 8 or 24; depth of windows to display in
       RETURNS : conn           - an open DX connection

  """

  assert hw_render_mode in [ "none", "gl","opengl" ]

  os.putenv( "DXNO_BACKING_STORE", "1" );     # Make any difference??
  os.putenv( "DXFLING"           , "1" );     # Rotate after release
  
  cmd = DX.DXEXEC + " -execonly"
  if hw_render_mode != "none": cmd = cmd + " -hwrender %s" % hw_render_mode
  if depth          == 8     : cmd = cmd + " -8bitcmap private"
  conn = DX.DXLStartDX( cmd, None )
  #DXLSetMessageDebugging( conn, 1 )

  return conn

#-----------------------------------------------------------------------------

def test():

  depth = 0

  # Parse arguments

  if len( sys.argv ) > 1:
    try:    depth = int( sys.argv[2] )
    except: depth = 0
    if depth not in [ 8, 24 ] or sys.argv[1] != "-depth":
      print "FORMAT: %s [-depth 8|24]\n" % sys.argv[0]
      sys.exit(0)
     
  # Determine the depth, if non specified
  root     = Tk()
  root.wm_withdraw()

  if not depth: depth = root.winfo_depth()

  if depth == 8: visual = "pseudocolor 8"
  else:          visual = "truecolor  24"

  # Create the GUI
  main     = Toplevel( root, visual=visual )
  frame    = Frame( main, visual=visual, 
                    width = IMAGE_WIDTH, height = IMAGE_HEIGHT )
  frame.pack_propagate( 0 )

  picture  = Frame( frame, background="", colormap="new" )
  button   = Button( main, text="Quit", command = root.quit )

  # FIXME:  Need to tell DX to resize when we resize
  picture.pack( expand=Y, fill=BOTH )
  frame.pack( expand=Y, fill=BOTH )
  button.pack()

  main.update_idletasks()

  # Display config
  localhost   = socket.gethostname()
  displayhost = string.split( root.winfo_screen(), ':' )[0]
  if not displayhost or displayhost == "unix":
    displayhost = localhost

  print "\nPYTHON-DX COLOR CUBE INTERACTION"
  print   "________________________________\n"
  print "CONFIGURATION:"
  print "  Local Machine          : %s" % localhost
  print "  Display                : %s" % displayhost
  print "  Hardware Acceleration  : %s" % HW_RENDER_MODE
  print "  Rendering Approximation: %s" % HW_RENDER_APPROX

  # Create a DX connection displaying a color cube in our Tk widget's X window
  conn  = InitDX( depth, HW_RENDER_MODE )
  box   = Box( conn, picture, HW_RENDER_MODE, HW_RENDER_APPROX )

  root.mainloop()

  DX.DXLExitDX( conn )

if __name__ == "__main__":
  test()
