#!/usr/bin/env python
#
#  11-supervise_net.py  -  This is 10-supervise_spin.py reworked so that the 
#                          network is imported from a .net file, for testing
#                          purposes.  The net is parameterized with DXLInput
#                          modules.   
#
#  NOTES:
#    1) Parameterizing values
#       a) In a UI net, you pop in DXLInputs which can be populated using
#          DXLSet... calls
#       b) In a DXLSend-created network, you simply use placeholder names
#          and re-set these values using DXLSend( 'var=val;' ) calls.
#          See below for an example.
#
#

import sys, os, math, time, socket, string, DX
from Tkinter import *

IMAGE_WIDTH      = 640
IMAGE_HEIGHT     = 480
HW_RENDER_MODE   = "opengl"                # "none", "gl", "opengl" 
#HW_RENDER_MODE   = "none"                # "none", "gl", "opengl" 
HW_RENDER_APPROX = "none"                 # "none", "box", "dots", "wireframe"

#-----------------------------------------------------------------------------

class Box:
  """  A colored box object displayed in DX.  """
  
  def __init__( self, conn, wgt, hw_render_mode, hw_render_approx ):
    """  Generate a colored box to use for display in DX.

         INPUT  :  conn             - open DX connection
                   wgt              - widget to tell DX to blast in
                   hw_render_mode   - "none", "gl", "opengl"
                   hw_render_approx - "none", "box", "dots", "wireframe"
                   
         RETURNS:  <string> - name of DX script language variable holding obj
    """

    assert hw_render_mode   in [ "none", "gl","opengl" ]
    assert hw_render_approx in [ "none", "box", "dots", "wireframe" ]

    self.conn             = conn
    self.wgt              = wgt
    self.hw_render_mode   = hw_render_mode
    self.hw_render_approx = hw_render_approx

    self.orient_v      = [ 1,0,0 ]
    self.up_v          = [ 0,1,0 ]

    # One area changed.  Instead of creating the network with DXLSend
    #   calls for each module, we just load a canned DX network from disk,
    #   and set up the DXLInput parameters with DXLSet... calls.
    DX.exDXLLoadScript( conn, os.path.join( _TEST_DIR, "11-supervise_net.net" ) )

    ( width, height, depth ) = ( wgt.winfo_width(), wgt.winfo_height(),
                                 wgt.winfo_depth() )

    if self.hw_render_mode == "none":  rendering_mode = "software"
    else:                              rendering_mode = "hardware"
    
    DX.DXLSetString ( conn, 'rendering_mode'        , rendering_mode              )
    DX.DXLSetString ( conn, 'rendering_appoximation', self.hw_render_approx       )
    DX.DXLSetInteger( conn, 'camera_resolution'     , width                       )
    DX.DXLSetScalar ( conn, 'camera_aspect'         , float( height ) / width     )
    DX.DXLSetValue  ( conn, 'supervisewin_size'     , '[%d %d]' % (width, height) )
    DX.DXLSetInteger( conn, 'supervisewin_parent'   , wgt.winfo_id()              )
    DX.DXLSetInteger( conn, 'supervisewin_depth'    , depth                       )

    self.SetOrientation( self.orient_v, self.up_v )

  def SetOrientation( self, orient_v, up_v ):

    self.orient_v = orient_v[:]
    self.up_v     = up_v    [:]

    def vector3_to_dx( v ):
      return "[ %g %g %g ]" % tuple(v)

    DX.DXLSetValue ( self.conn, 'camera_direction',
                  vector3_to_dx( self.orient_v ) )
    DX.DXLSetValue ( self.conn, 'camera_up'       ,
                  vector3_to_dx( self.up_v )     )

  def Display( self ):

    DX.DXLExecuteOnce( self.conn )
    DX.WaitForDXIdle( self.conn )

#-----------------------------------------------------------------------------

def InitDX( depth, hw_render_mode ):
  """  Initializes and returns a connection to the DX executive.

       INPUT   : hw_render_mode - "none", "gl", "opengl"
                 depth          - 8 or 24; depth of windows to display in
       RETURNS : conn           - an open DX connection

  """

  assert hw_render_mode in [ "none", "gl","opengl" ]

  os.putenv( "DXNO_BACKING_STORE", "1" );     # Make any difference??
  os.putenv( "DXFLING"           , "1" );     # Rotate after release
  
  cmd = DX.DXEXEC + " -execonly"
  if hw_render_mode != "none": cmd = cmd + " -hwrender %s" % hw_render_mode
  if depth          == 8     : cmd = cmd + " -8bitcmap private"
  conn = DX.DXLStartDX( cmd, None )
#  DX.DXLSetMessageDebugging( conn, 1 )

  return conn

#-----------------------------------------------------------------------------

def test():

  depth = 0

  # Parse arguments

  if len( sys.argv ) > 1:
    try:    depth = int( sys.argv[2] )
    except: depth = 0
    if depth not in [ 8, 24 ] or sys.argv[1] != "-depth":
      print "FORMAT: %s [-depth 8|24]\n" % sys.argv[0]
      sys.exit(0)
     
  # Determine the depth, if non specified
  root     = Tk()
  root.wm_withdraw()

  if not depth: depth = root.winfo_depth()

  if depth == 8: visual = "pseudocolor 8"
  else:          visual = "truecolor  24"

  # Create the GUI
  main     = Toplevel( root, visual=visual )
  frame    = Frame( main, visual=visual, 
                    width = IMAGE_WIDTH, height = IMAGE_HEIGHT )
  frame.pack_propagate( 0 )

  picture  = Frame( frame, background="", colormap="new" )
  button   = Button( main, text="Quit", command = root.quit )

  # FIXME:  Need to tell DX to resize when we resize
  picture.pack( expand=Y, fill=BOTH )
  frame.pack( expand=Y, fill=BOTH )
  button.pack()

  main.update_idletasks()

  # Display config
  localhost   = socket.gethostname()
  displayhost = string.split( root.winfo_screen(), ':' )[0]
  if not displayhost or displayhost == "unix":
    displayhost = localhost

  print "\nPYTHON-DX COLOR CUBE ROTATE DEMO"
  print   "________________________________\n"
  print "CONFIGURATION:"
  print "  Local Machine          : %s" % localhost
  print "  Display                : %s" % displayhost
  print "  Hardware Acceleration  : %s" % HW_RENDER_MODE
  print "  Rendering Approximation: %s" % HW_RENDER_APPROX

  # Create a DX connection displaying a color cube inside our Tk widget's window
  conn  = InitDX( depth, HW_RENDER_MODE )
  box   = Box( conn, picture, HW_RENDER_MODE, HW_RENDER_APPROX )

  time1      = time.time()
  num_frames = 0
  
  for i in range(0,360,2):
    box.SetOrientation( [ math.cos( i * math.pi / 180.0 ), 0.5,
                          math.sin( i * math.pi / 180.0 ) ],
                        [0,1,0] )
    box.Display()
    num_frames = num_frames + 1

  time2 = time.time()
  print "Done ... %g frames/sec" % ( num_frames / (time2 - time1) )

  root.mainloop()

  DX.DXLExitDX( conn )

if __name__ == "__main__":

  global _TEST_DIR
  _TEST_DIR = os.path.abspath( sys.path[0] or os.curdir )
  
  test()
